@@ -24,43 +24,79 @@ const Gtk = imports.gi.Gtk;
 const _ = imports.gettext.gettext;
 
 const Lang = imports.lang;
+const Mainloop = imports.mainloop;
 const Signals = imports.signals;
 
 const Documents = imports.documents;
 const Global = imports.global;
 const TrackerUtils = imports.trackerUtils;
+const WindowMode = imports.windowMode;
 const Utils = imports.utils;
 
-function ContextMenu(urn) {
-    this._init(urn);
+function ContextMenu(urns) {
+    this._init(urns);
 }
 
 ContextMenu.prototype = {
-    _init: function(urn) {
-        let doc = Global.documentManager.lookupDocument(urn);
-        let isFavorite = doc.favorite;
+    _init: function(urns) {
+        let favCount = 0;
+        let apps = [];
+        let docs = [];
 
         this.widget = new Gtk.Menu();
+        let showFavorite = (Global.modeController.getWindowMode() != WindowMode.WindowMode.PREVIEW);
+
+        urns.forEach(Lang.bind(this,
+            function(urn) {
+                let doc = Global.documentManager.lookupDocument(urn);
+                if (doc.favorite)
+                    favCount++;
+
+                if (apps.indexOf(doc.defaultAppName) == -1) {
+                    apps.push(doc.defaultAppName);
+                }
+
+                docs.push(doc);
+            }));
+
+        showFavorite &= ((favCount == 0) || (favCount == urns.length));
+
+        let openLabel = null;
+        if (apps.length == 1) {
+            // Translators: this is the Open action in a context menu
+            openLabel = _("Open with %s").format(apps[0]);
+        } else {
+            // Translators: this is the Open action in a context menu
+            openLabel = _("Open");
+        }
 
-        // Translators: this is the Open action in a context menu
-        let openLabel = (doc.defaultAppName) ? _("Open with %s").format(doc.defaultAppName) : _("Open");
         let openItem = new Gtk.MenuItem({ label: openLabel });
         openItem.show();
         this.widget.append(openItem);
 
-        openItem.connect('activate', Lang.bind(this,
-            function(item) {
-                doc.open(item.get_screen(), Gtk.get_current_event_time());
-            }));
+        let favoriteItem = null;
+        if (showFavorite) {
+            let isFavorite = (favCount == urns.length);
+            let favoriteLabel = (isFavorite) ? _("Remove from favorites") : _("Add to favorites");
 
-        let favoriteLabel = (isFavorite) ? _("Remove from favorites") : _("Add to favorites");
-        let favoriteItem = new Gtk.MenuItem({ label: favoriteLabel });
-        favoriteItem.show();
-        this.widget.append(favoriteItem);
+            favoriteItem = new Gtk.MenuItem({ label: favoriteLabel });
+            favoriteItem.show();
+            this.widget.append(favoriteItem);
+        }
 
-        favoriteItem.connect('activate', Lang.bind(this,
-            function() {
-                doc.setFavorite(!isFavorite);
+        docs.forEach(Lang.bind(this,
+            function(doc) {
+                openItem.connect('activate', Lang.bind(this,
+                    function(item) {
+                        doc.open(item.get_screen(), Gtk.get_current_event_time());
+                    }));
+
+                if (favoriteItem) {
+                    favoriteItem.connect('activate', Lang.bind(this,
+                        function() {
+                            doc.setFavorite(!doc.favorite);
+                        }));
+                }
             }));
 
         this.widget.show_all();
@@ -75,15 +111,15 @@ View.prototype = {
     _init: function() {
         this._selectedURNs = null;
 
-        this._model = new Documents.DocumentModel();
-        this._treeModel = this._model.model;
+        this._treeModel = Global.documentManager.getModel().model;
         this.widget.set_model(this._treeModel);
 
         this.widget.connect('destroy', Lang.bind(this,
             function() {
                 Global.selectionController.disconnect(this._selectionControllerId);
             }));
-        this.widget.connect('button-release-event', Lang.bind(this, this._onButtonRelease));
+        this.widget.connect('button-release-event', Lang.bind(this, this._onButtonReleaseEvent));
+        this.widget.connect('button-press-event', Lang.bind(this, this._onButtonPressEvent));
 
         this.createRenderers();
 
@@ -92,12 +128,19 @@ View.prototype = {
             this._selectionController.connect('selection-check',
                                               Lang.bind(this, this._updateSelection));
 
-        this._updateSelection(true);
+        // HACK: give the view some time to setup the scrolled window
+        // allocation, as updateSelection() might call scrollToPath().
+        // Is there anything better we can do here?
+        Mainloop.timeout_add(100, Lang.bind(this,
+            function() {
+                this._updateSelection();
+                return false;
+            }));
 
         this.connectToSelectionChanged(Lang.bind(this, this._onSelectionChanged));
     },
 
-    _updateSelection: function(scroll) {
+    _updateSelection: function() {
         let selectionObject = this.getSelectionObject();
         let selected = this._selectionController.getSelection().slice(0);
 
@@ -114,7 +157,7 @@ View.prototype = {
                     selectionObject.select_path(path);
                     selected.splice(urnIndex, 1);
 
-                    if (first && scroll) {
+                    if (first) {
                         this.scrollToPath(path);
                         first = false;
                     }
@@ -133,7 +176,29 @@ View.prototype = {
         Global.selectionController.setSelection(selectedURNs);
     },
 
-    _onButtonRelease: function(view, event) {
+    _onButtonPressEvent: function(widget, event) {
+        let button = event.get_button()[1];
+        if (button != 3)
+            return false;
+
+        let coords = [ event.get_coords()[1] , event.get_coords()[2] ];
+        let path = this.getPathAtPos(coords);
+
+        let selection = Global.selectionController.getSelection();
+
+        if (path) {
+            let urn = Utils.getURNFromPath(path, this._treeModel);
+
+            if (selection.indexOf(urn) == -1)
+                this.getSelectionObject().unselect_all();
+
+            this.getSelectionObject().select_path(path);
+        }
+
+        return true;
+    },
+
+    _onButtonReleaseEvent: function(view, event) {
         let button = event.get_button()[1];
         let coords = [ event.get_coords()[1] , event.get_coords()[2] ];
         let timestamp = event.get_time();
@@ -142,10 +207,14 @@ View.prototype = {
             return false;
 
         let path = this.getPathAtPos(coords);
+
+        if (!path)
+            return false;
+
         let iter = this._treeModel.get_iter(path)[1];
 
         let urn = this._treeModel.get_value(iter, Documents.ModelColumns.URN);
-        let menu = new ContextMenu(urn);
+        let menu = new ContextMenu(Global.selectionController.getSelection());
 
         menu.widget.popup_for_device(null, null, null, null, null, null, button, timestamp);
 
